import dataclasses
import enum
import logging
import re
from typing import List, Optional, Tuple, Union

from tavern._core import exceptions
from tavern._core.strtobool import strtobool

logger: logging.Logger = logging.getLogger(__name__)


class StrictSetting(enum.Enum):
    """The actual setting for a particular block"""

    ON = 1
    OFF = 2
    UNSET = 3
    LIST_ANY_ORDER = 4


valid_keys = ["json", "headers", "redirect_query_params"]

valid_switches = ["on", "off", "list_any_order"]


def strict_setting_factory(str_setting: Optional[str]) -> StrictSetting:
    """Converts from cmdline/setting file to an enum"""
    if str_setting is None:
        return StrictSetting.UNSET
    else:
        if str_setting == "list_any_order":
            return StrictSetting.LIST_ANY_ORDER

        parsed = strtobool(str_setting)

        if parsed:
            return StrictSetting.ON
        else:
            return StrictSetting.OFF


@dataclasses.dataclass(frozen=True)
class StrictOption:
    """The section and the setting. The setting is only stored here because json works slightly
    differently, otherwise it's redundant"""

    section: str
    setting: StrictSetting

    def is_on(self) -> bool:
        if self.section == "json":
            # Must be specifically disabled for response body
            return self.setting not in [StrictSetting.OFF, StrictSetting.LIST_ANY_ORDER]
        else:
            # Off by default for everything else
            return self.setting in [StrictSetting.ON]


def validate_and_parse_option(key: str) -> StrictOption:
    regex = re.compile(
        "(?P<section>{sections})(:(?P<setting>{switches}))?".format(
            sections="|".join(valid_keys), switches="|".join(valid_switches)
        )
    )

    match = regex.fullmatch(key)

    if not match:
        raise exceptions.InvalidConfigurationException(
            "Invalid value for 'strict' given - expected one of {}, got '{}'".format(
                [f"{key}[:on/off]" for key in valid_keys], key
            )
        )

    as_dict = match.groupdict()

    if as_dict["section"] != "json" and as_dict["setting"] == "list_any_order":
        logger.warning(
            "Using 'list_any_order' key outside of 'json' section has no meaning"
        )

    return StrictOption(as_dict["section"], strict_setting_factory(as_dict["setting"]))


@dataclasses.dataclass(frozen=True)
class StrictLevel:
    """Strictness settings for every block in a response

    TODO: change the name of this class, it's awful"""

    json: StrictOption = dataclasses.field(
        default=StrictOption("json", strict_setting_factory(None))
    )
    headers: StrictOption = dataclasses.field(
        default=StrictOption("headers", strict_setting_factory(None))
    )
    redirect_query_params: StrictOption = dataclasses.field(
        default=StrictOption("redirect_query_params", strict_setting_factory(None))
    )

    @classmethod
    def from_options(cls, options: Union[List[str], str]) -> "StrictLevel":
        if isinstance(options, str):
            options = [options]
        elif not isinstance(options, list):
            raise exceptions.InvalidConfigurationException(
                "'strict' setting should be a list of strings"
            )

        logger.debug("Parsing options to strict level: %s", options)

        parsed = [validate_and_parse_option(key) for key in options]

        return cls(**{i.section: i for i in parsed})

    def option_for(self, section: str) -> StrictOption:
        """Provides a string-based way of getting strict settings for a section"""
        try:
            return getattr(self, section)
        except AttributeError as e:
            raise exceptions.InvalidConfigurationException(
                f"No setting for '{section}'"
            ) from e

    @classmethod
    def all_on(cls) -> "StrictLevel":
        return cls.from_options([i + ":on" for i in valid_keys])

    @classmethod
    def all_off(cls) -> "StrictLevel":
        return cls.from_options([i + ":off" for i in valid_keys])


StrictSettingKinds = Union[None, bool, StrictSetting, StrictOption]


def extract_strict_setting(strict: StrictSettingKinds) -> Tuple[bool, StrictSetting]:
    """Takes either a bool, StrictOption, or a StrictSetting and return the bool representation
    and StrictSetting representation"""

    logger.debug("Parsing a '%s': %s", type(strict), strict)

    if isinstance(strict, StrictSetting):
        strict_setting = strict
        strict = strict == StrictSetting.ON
    elif isinstance(strict, StrictOption):
        strict_setting = strict.setting
        strict = strict.is_on()
    elif isinstance(strict, bool):
        strict_setting = strict_setting_factory(str(strict))
    elif strict is None:
        strict = False
        strict_setting = strict_setting_factory("false")
    else:
        raise exceptions.InvalidConfigurationException(
            "Unable to parse strict setting '{}' of type '{}'".format(
                strict, type(strict)
            )
        )

    logger.debug("Got strict as '%s', setting as '%s'", strict, strict_setting)

    return strict, strict_setting



import copy
import functools
import itertools
import logging
import typing
from typing import Any, Callable, Dict, Iterable, Iterator, List, Mapping, Tuple, Union

import pytest
import yaml
from box import Box
from pytest import Mark

from tavern._core import exceptions
from tavern._core.dict_util import deep_dict_merge, format_keys, get_tavern_box
from tavern._core.extfunctions import get_wrapped_create_function, is_ext_function
from tavern._core.loader import IncludeLoader
from tavern._core.schema.files import verify_tests

from .item import YamlItem
from .util import load_global_cfg

logger: logging.Logger = logging.getLogger(__name__)

T = typing.TypeVar("T")

_format_without_inner: Callable[[T, Mapping], T] = functools.partial(
    format_keys, no_double_format=False
)


def _format_test_marks(
    original_marks: Iterable[Union[str, dict]], fmt_vars: Mapping, test_name: str
) -> Tuple[List[Mark], List[Mapping]]:
    """Given the 'raw' marks from the test and any available format variables,
    generate new  marks for this test

    Args:
        original_marks: Raw string from test - should correspond to either a
            pytest builtin mark or a custom user mark
        fmt_vars: dictionary containing available format variables
        test_name: Name of test (for error logging)

    Returns:
        first element is normal pytest mark objects, second element is all
            marks which were formatted (no matter their content)

    Todo:
        Fix doctests below - failing due to missing pytest markers

    Example:

        # >>> _format_test_marks([], {}, 'abc')
        # ([], [])
        # >>> _format_test_marks(['tavernmarker'], {}, 'abc')
        # (['tavernmarker'], [])
        # >>> _format_test_marks(['{formatme}'], {'formatme': 'tavernmarker'}, 'abc')
        # (['tavernmarker'], [])
        # >>> _format_test_marks([{'skipif': '{skiptest}'}], {'skiptest': true}, 'abc')
        # (['tavernmarker'], [])

    """

    pytest_marks: List[Mark] = []
    formatted_marks: List[Mapping] = []

    for m in original_marks:
        if isinstance(m, str):
            # a normal mark
            m = _format_without_inner(m, fmt_vars)
            pytest_marks.append(getattr(pytest.mark, m))
        elif isinstance(m, dict):
            # skipif or parametrize (for now)
            for markname, extra_arg in m.items():
                # NOTE
                # cannot do 'skipif' and rely on a parametrized
                # argument.
                try:
                    extra_arg = _format_without_inner(extra_arg, fmt_vars)
                except exceptions.MissingFormatError as e:
                    msg = "Tried to use mark '{}' (with value '{}') in test '{}' but one or more format variables was not in any configuration file used by the test".format(
                        markname, extra_arg, test_name
                    )
                    # NOTE
                    # we could continue and let it fail in the test, but
                    # this gives a better indication of what actually
                    # happened (even if it is difficult to test)
                    raise exceptions.MissingFormatError(msg) from e
                else:
                    pytest_marks.append(getattr(pytest.mark, markname)(extra_arg))
                    formatted_marks.append({markname: extra_arg})
        else:
            raise exceptions.BadSchemaError(f"Unexpected mark type '{type(m)}'")

    return pytest_marks, formatted_marks


def _maybe_load_ext(pair):
    """Try to load ext values"""
    key, value = pair

    if is_ext_function(value):
        # If it is an ext function, load the new (or supplemental) value[s]
        ext = value.pop("$ext")
        f = get_wrapped_create_function(ext)
        new_value = f()

        if len(value) == 0:
            # Use only this new value
            return key, new_value
        elif isinstance(new_value, dict):
            # Merge with some existing data. At this point 'value' is known to be a dict.
            return key, deep_dict_merge(value, f())
        else:
            # For example, if it's defined like
            #
            # - testkey: testval
            #   $ext:
            #     function: mod:func
            #
            # and 'mod:func' returns a string, it's impossible to 'merge' with the existing data.
            logger.error("Values still in 'val': %s", value)
            raise exceptions.BadSchemaError(
                "There were extra key/value pairs in the 'val' for this parametrize mark, but the ext function {} returned '{}' (of type {}) that was not a dictionary. It is impossible to merge these values.".format(
                    ext, new_value, type(new_value)
                )
            )

    return key, value


def _generate_parametrized_test_items(
    keys: Iterable[Union[str, List, Tuple]], vals_combination: Iterable[Tuple[str, str]]
) -> Tuple[Mapping[str, Any], str]:
    """Generate test name from given key(s)/value(s) combination

    Args:
        keys: list of keys to format name with
        vals_combination this combination of values for the key

    Returns:
        tuple of the variables for the stage and the generated stage name
    """
    flattened_values: List[Iterable[str]] = []
    variables: Dict[str, Any] = {}

    # combination of keys and the values they correspond to
    for pair in zip(keys, vals_combination):
        key, value = pair
        # NOTE: If test is invalid, test names generated here will be
        # very weird looking
        if isinstance(key, str):
            variables[key] = value
            flattened_values.append(value)
        else:
            if not isinstance(value, (list, tuple)):
                value = [value]

            if len(value) != len(key):
                raise exceptions.BadSchemaError(
                    "Invalid match between numbers of keys and number of values in parametrize mark ({} keys, {} values)".format(
                        key, value
                    )
                )

            for subkey, subvalue in zip(key, value):
                variables[subkey] = subvalue
                flattened_values.append(subvalue)

    variables = dict(map(_maybe_load_ext, variables.items()))

    logger.debug("Variables for this combination: %s", variables)
    logger.debug("Values for this combination: %s", flattened_values)

    # Use for formatting parametrized values - eg {}-{}, {}-{}-{}, etc.
    inner_fmt = "-".join(["{}"] * len(flattened_values))
    inner_formatted = inner_fmt.format(*flattened_values)

    return variables, inner_formatted


def _get_parametrized_items(
    parent: pytest.File,
    test_spec: Dict,
    parametrize_marks: List[Dict],
    pytest_marks: List[pytest.Mark],
) -> Iterator[YamlItem]:
    """Return new items with new format values available based on the mark

    This will change the name from something like 'test a thing' to 'test a
    thing[param1]', 'test a thing[param2]', etc. This probably messes with
    -k

    Note:
        This still has the pytest.mark.parametrize mark on it, though it
        doesn't appear to do anything. This could be removed?
    """

    logger.debug("parametrize marks: %s", parametrize_marks)

    # These should be in the same order as specified in the input file
    vals = [i["parametrize"]["vals"] for i in parametrize_marks]

    logger.debug("(possibly wrapped) values: %s", vals)

    def unwrap_map(value):
        if is_ext_function(value):
            ext = value.pop("$ext")
            f = get_wrapped_create_function(ext)
            new_value = f()
            return new_value

        return value

    vals = list(map(unwrap_map, vals))

    try:
        combined = itertools.product(*vals)
    except TypeError as e:
        raise exceptions.BadSchemaError(
            "Invalid match between numbers of keys and number of values in parametrize mark"
        ) from e

    keys: List[str] = [i["parametrize"]["key"] for i in parametrize_marks]

    for vals_combination in combined:
        logger.debug("Generating test for %s/%s", keys, vals_combination)

        if len(vals_combination) != len(keys):
            raise exceptions.BadSchemaError(
                "Invalid match between numbers of keys and number of values in parametrize mark"
            )

        variables, inner_formatted = _generate_parametrized_test_items(
            keys, vals_combination
        )

        # Change the name
        spec_new = copy.deepcopy(test_spec)
        spec_new["test_name"] = test_spec["test_name"] + f"[{inner_formatted}]"

        logger.debug("New test name: %s", spec_new["test_name"])

        # Make this new thing available for formatting
        spec_new.setdefault("includes", []).append(
            {
                "name": f"parametrized[{inner_formatted}]",
                "description": "autogenerated by Tavern",
                "variables": variables,
            }
        )
        # And create the new item
        item_new = YamlItem.yamlitem_from_parent(
            spec_new["test_name"], parent, spec_new, parent.path
        )
        item_new.add_markers(pytest_marks)

        yield item_new


class YamlFile(pytest.File):
    """Custom `File` class that loads each test block as a different test"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # This (and the FakeObj below) are to make pytest-pspec not error out.
        # The 'docstring' for this is the filename, the 'docstring' for each
        # individual test is the actual test name.
        class FakeObj:
            __doc__ = str(self.path)

        self.obj = FakeObj

    def _get_test_fmt_vars(self, test_spec: Mapping) -> dict:
        """Get any format variables that can be inferred for the test at this point

        Args:
            test_spec: Test specification, possibly with included config files

        Returns:
            available format variables
        """
        # Get included variables so we can do things like:
        # skipif: {my_integer} > 2
        # skipif: 'https' in '{hostname}'
        # skipif: '{hostname}'.contains('ignoreme')
        fmt_vars: Dict = {}

        global_cfg = load_global_cfg(self.config)
        fmt_vars.update(**global_cfg.variables)

        included = test_spec.get("includes", [])
        for i in included:
            fmt_vars.update(**i.get("variables", {}))

        if self.session.config.option.collectonly:
            tavern_box = Box(default_box=True)
        else:
            # Needed if something in a config file uses tavern.env_vars
            tavern_box = get_tavern_box()

        try:
            fmt_vars = _format_without_inner(fmt_vars, tavern_box)
        except exceptions.MissingFormatError as e:
            # eg, if we have {tavern.env_vars.DOESNT_EXIST}
            msg = "Tried to use tavern format variable that did not exist"
            raise exceptions.MissingFormatError(msg) from e

        tavern_box.merge_update(**fmt_vars)
        return tavern_box

    def _generate_items(self, test_spec: Dict) -> Iterator[YamlItem]:
        """Modify or generate tests based on test spec

        If there are any 'parametrize' marks, this will generate extra tests
        based on the values

        Args:
            test_spec: Test specification

        Yields:
            Tavern YAML test
        """
        item = YamlItem.yamlitem_from_parent(
            test_spec["test_name"], self, test_spec, self.path
        )

        original_marks = test_spec.get("marks", [])

        if original_marks:
            fmt_vars = self._get_test_fmt_vars(test_spec)
            pytest_marks, formatted_marks = _format_test_marks(
                original_marks, fmt_vars, test_spec["test_name"]
            )

            # Do this after we've added all the other marks so doing
            # things like selecting on mark names still works even
            # after parametrization
            parametrize_marks = [
                i for i in formatted_marks if isinstance(i, dict) and "parametrize" in i
            ]

            if parametrize_marks:
                yield from _get_parametrized_items(
                    self, test_spec, parametrize_marks, pytest_marks
                )

                # Only yield the parametrized ones
                return
            else:
                item.add_markers(pytest_marks)

        yield item

    def collect(self) -> Iterator[YamlItem]:
        """Load each document in the given input file into a different test

        Yields:
            Pytest 'test objects'
        """

        try:
            # Convert to a list so we can catch parser exceptions
            all_tests: Iterable[dict] = list(
                yaml.load_all(
                    self.path.open(encoding="utf-8"),
                    Loader=IncludeLoader,  # type:ignore
                )
            )
        except yaml.parser.ParserError as e:
            raise exceptions.BadSchemaError from e

        for test_spec in all_tests:
            if not test_spec:
                logger.warning("Empty document in input file '%s'", self.path)
                continue

            try:
                for i in self._generate_items(test_spec):
                    i.initialise_fixture_attrs()
                    yield i
            except (TypeError, KeyError) as e:
                try:
                    verify_tests(test_spec, with_plugins=False)
                except Exception as e2:
                    raise e2 from e
                else:
                    raise



import logging
import pathlib
from typing import MutableMapping, Optional, Tuple, Union

import attr
import pytest
import yaml
from _pytest._code.code import ExceptionInfo, TerminalRepr
from _pytest.nodes import Node

from tavern._core import exceptions
from tavern._core.loader import error_on_empty_scalar
from tavern._core.plugins import load_plugins
from tavern._core.pytest import call_hook
from tavern._core.pytest.error import ReprdError
from tavern._core.report import attach_text
from tavern._core.run import run_test
from tavern._core.schema.files import verify_tests

from .config import TestConfig
from .util import load_global_cfg

logger: logging.Logger = logging.getLogger(__name__)


class YamlItem(pytest.Item):
    """Simple wrapper around new test type that can report errors more
    accurately than the default pytest reporting stuff

    At the time of writing this doesn't print the error very nicely, but it
    should be enough to track down what went wrong

    Attributes:
        path: filename that this test came from
        spec: The whole dictionary of the test
        global_cfg: configuration for test
    """

    # See https://github.com/taverntesting/tavern/issues/825
    _patched_yaml = False

    global_cfg: TestConfig

    def __init__(
        self, *, name: str, parent, spec: MutableMapping, path: pathlib.Path, **kwargs
    ) -> None:
        if "grpc" in spec:
            logger.warning("Tavern grpc support is in an experimental stage")

        super().__init__(name, parent, **kwargs)
        self.path = path
        self.spec = spec

        if not YamlItem._patched_yaml:
            yaml.parser.Parser.process_empty_scalar = (  # type:ignore
                error_on_empty_scalar
            )

            YamlItem._patched_yaml = True

    @classmethod
    def yamlitem_from_parent(cls, name, parent: Node, spec, path: pathlib.Path):
        return cls.from_parent(parent, name=name, spec=spec, path=path)

    def initialise_fixture_attrs(self) -> None:
        self.funcargs = {}  # type: ignore

        # _get_direct_parametrize_args checks parametrize arguments in Python
        # functions, but we don't care about that in Tavern.
        self.session._fixturemanager._get_direct_parametrize_args = (  # type: ignore
            lambda _: []  # type: ignore
        )  # type: ignore

        fixtureinfo = self.session._fixturemanager.getfixtureinfo(
            self, self.obj, type(self), funcargs=False
        )
        self._fixtureinfo = fixtureinfo
        self.fixturenames = fixtureinfo.names_closure
        self._request = pytest.FixtureRequest(self, _ispytest=True)

    @property
    def location(self):
        """get location in file"""
        location = super().location
        location = (location[0], self.spec.start_mark.line, location[2])
        return location

    #     Hack to stop issue with pytest-rerunfailures
    _initrequest = initialise_fixture_attrs

    def setup(self) -> None:
        super().setup()
        self._request._fillfixtures()

    @property
    def obj(self):
        stages = []
        for i, stage in enumerate(self.spec["stages"]):
            name = "<unknown>"
            if "name" in stage:
                name = stage["name"]
            elif "id" in stage:
                name = stage["id"]
            stages.append(f"{i + 1:d}: {name:s}")

        # This needs to be a function or skipif breaks
        def fakefun():
            pass

        fakefun.__doc__ = self.name + ":\n" + "\n".join(stages)
        return fakefun

    @property
    def _obj(self):
        return self.obj

    def add_markers(self, pytest_marks) -> None:
        for pm in pytest_marks:
            if pm.name == "usefixtures":
                if (
                    not isinstance(pm.mark.args, (list, tuple))
                    or len(pm.mark.args) == 0
                ):
                    logger.error(
                        "'usefixtures' was an invalid type (should"
                        " be a list of fixture names)"
                    )
                    continue
                # Need to do this here because we expect a list of markers from
                # usefixtures, which pytest then wraps in a tuple. we need to
                # extract this tuple so pytest can use both fixtures.
                if isinstance(pm.mark.args[0], (list, tuple)):
                    new_mark = attr.evolve(pm.mark, args=pm.mark.args[0])
                    pm = attr.evolve(pm, mark=new_mark)
                elif isinstance(pm.mark.args[0], (dict)):
                    # We could raise a TypeError here instead, but then it's a
                    # failure at collection time (which is a bit annoying to
                    # deal with). Instead just don't add the marker and it will
                    # raise an exception at test verification.
                    logger.error(
                        "'usefixtures' was an invalid type (should"
                        " be a list of fixture names)"
                    )
                    continue

            self.add_marker(pm)

    def _load_fixture_values(self):
        fixture_markers = self.iter_markers("usefixtures")

        values = {}

        for m in fixture_markers:
            if isinstance(m.args, (list, tuple)):
                mark_values = {f: self.funcargs[f] for f in m.args}
            elif isinstance(m.args, str):
                # Not sure if this can happen if validation is working
                # correctly, but it appears to be slightly broken so putting
                # this check here just in case
                mark_values = {m.args: self.funcargs[m.args]}
            else:
                raise exceptions.BadSchemaError(
                    (
                        "Can't handle 'usefixtures' spec of '{}'."
                        " There appears to be a bug in pykwalify so verification of"
                        " 'usefixtures' is broken - it should be a list of fixture"
                        " names"
                    ).format(m.args)
                )

            if any(mv in values for mv in mark_values):
                logger.warning("Overriding value for %s", mark_values)

            values.update(mark_values)

        # Use autouse fixtures as well
        for name in self.fixturenames:
            if name in values:
                logger.debug("%s already explicitly used", name)
                continue

            mark_values = {name: self.funcargs[name]}
            values.update(mark_values)

        return values

    def runtest(self) -> None:
        self.global_cfg = load_global_cfg(self.config)

        load_plugins(self.global_cfg)

        # INTERNAL
        xfail = self.spec.get("_xfail", False)

        try:
            fixture_values = self._load_fixture_values()
            self.global_cfg.variables.update(fixture_values)

            call_hook(
                self.global_cfg,
                "pytest_tavern_beta_before_every_test_run",
                test_dict=self.spec,
                variables=self.global_cfg.variables,
            )

            verify_tests(self.spec)

            for stage in self.spec["stages"]:
                if not stage.get("name"):
                    if not stage.get("id"):
                        # Should never actually reach here, should be caught at schema check time
                        raise exceptions.BadSchemaError(
                            "One of name or ID must be specified"
                        )

                    stage["name"] = stage["id"]

            run_test(self.path, self.spec, self.global_cfg)

        except exceptions.BadSchemaError:
            if xfail == "verify":
                logger.info("xfailing test while verifying schema")
                self.add_marker(pytest.mark.xfail, True)
            raise
        except exceptions.TavernException as e:
            if isinstance(xfail, dict):
                if msg := xfail.get("run"):
                    if msg not in str(e):
                        raise Exception(
                            f"error message did not match: expected '{msg}', got '{e!s}'"
                        ) from e
                    logger.info("xfailing test when running")
                    self.add_marker(pytest.mark.xfail, True)
                else:
                    logger.warning("internal error checking 'xfail'")
            elif xfail == "run" and not e.is_final:
                logger.info("xfailing test when running")
                self.add_marker(pytest.mark.xfail, True)
            elif xfail == "finally" and e.is_final:
                logger.info("xfailing test when finalising")
                self.add_marker(pytest.mark.xfail, True)

            raise
        else:
            if xfail:
                raise Exception(f"internal: xfail test did not fail '{xfail}'")
        finally:
            call_hook(
                self.global_cfg,
                "pytest_tavern_beta_after_every_test_run",
                test_dict=self.spec,
                variables=self.global_cfg.variables,
            )

    def repr_failure(
        self, excinfo: ExceptionInfo[BaseException], style: Optional[str] = None
    ) -> Union[TerminalRepr, str, ReprdError]:
        """called when self.runtest() raises an exception.

        By default, will raise a custom formatted traceback if it's a tavern error. if not, will use the default
        python traceback
        """

        if (
            self.config.getini("tavern-use-default-traceback")
            or self.config.getoption("tavern_use_default_traceback")
            or not issubclass(excinfo.type, exceptions.TavernException)
            or issubclass(excinfo.type, exceptions.BadSchemaError)
        ):
            return super().repr_failure(excinfo)

        if style is not None:
            logger.info("Ignoring style '%s", style)

        error = ReprdError(excinfo, self)
        attach_text(str(error), name="error_output")
        return error

    def reportinfo(self) -> Tuple[pathlib.Path, int, str]:
        return (
            self.path,
            0,
            "{s.path}::{s.name:s}".format(s=self),
        )



import copy
import dataclasses
import json
import os
import uuid
from copy import deepcopy
from unittest.mock import MagicMock, Mock, patch

import paho.mqtt.client as paho
import pytest
import requests

from tavern._core import exceptions
from tavern._core.pytest.util import load_global_cfg
from tavern._core.run import run_test
from tavern._plugins.mqtt.client import MQTTClient


@pytest.fixture(name="fulltest")
def fix_example_test():
    spec = {
        "test_name": "A test with a single stage",
        "stages": [
            {
                "name": "step 1",
                "request": {"url": "http://www.google.com", "method": "GET"},
                "response": {
                    "status_code": 200,
                    "json": {"key": "value"},
                    "headers": {"content-type": "application/json"},
                },
            }
        ],
    }

    return spec


@pytest.fixture(name="mockargs")
def fix_mock_response_args(fulltest):
    response = copy.deepcopy(fulltest["stages"][0]["response"])
    content = response["json"]

    args = {
        "spec": requests.Response,
        "content": json.dumps(content).encode("utf8"),
        "status_code": response["status_code"],
        "json": lambda: content,
        "headers": response["headers"],
    }

    return args


class TestRunStages:
    def test_success(self, fulltest, mockargs, includes):
        """Successful test"""

        mock_response = Mock(**mockargs)

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            return_value=mock_response,
        ) as pmock:
            run_test("heif", fulltest, includes)

        assert pmock.called

    def test_invalid_code(self, fulltest, mockargs, includes):
        """Wrong status code"""

        mockargs["status_code"] = 400

        mock_response = Mock(**mockargs)

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            return_value=mock_response,
        ) as pmock:
            with pytest.raises(exceptions.TestFailError):
                run_test("heif", fulltest, includes)

        assert pmock.called

    def test_invalid_body(self, fulltest, mockargs, includes):
        """Wrong body returned"""

        mockargs["json"] = lambda: {"wrong": "thing"}

        mock_response = Mock(**mockargs)

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            return_value=mock_response,
        ) as pmock:
            with pytest.raises(exceptions.TestFailError):
                run_test("heif", fulltest, includes)

        assert pmock.called

    def test_invalid_headers(self, fulltest, mockargs, includes):
        """Wrong headers"""

        mockargs["headers"] = {"content-type": "application/x-www-url-formencoded"}

        mock_response = Mock(**mockargs)

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            return_value=mock_response,
        ) as pmock:
            with pytest.raises(exceptions.TestFailError):
                run_test("heif", fulltest, includes)

        assert pmock.called


class TestIncludeStages:
    @pytest.fixture
    def fake_stages(self):
        stages = [
            {
                "id": "my_external_stage",
                "name": "My external stage",
                "request": {"url": "http://www.bing.com", "method": "GET"},
                "response": {
                    "status_code": 200,
                    "json": {"key": "value"},
                    "headers": {"content-type": "application/json"},
                },
            }
        ]

        return stages

    def check_mocks_called(self, pmock):
        assert pmock.called

        # We expect 2 calls, first to bing (external stage),
        # then google (part of fulltest)
        assert len(pmock.call_args_list) == 2
        args, kwargs = pmock.call_args_list[0]
        assert kwargs["url"] == "http://www.bing.com"
        args, kwargs = pmock.call_args_list[1]
        assert kwargs["url"] == "http://www.google.com"

    def test_included_stage(self, fulltest, mockargs, includes, fake_stages):
        """Load stage from includes"""
        mock_response = Mock(**mockargs)

        stage_includes = [{"stages": fake_stages}]

        newtest = deepcopy(fulltest)
        newtest["includes"] = stage_includes
        newtest["stages"].insert(0, {"type": "ref", "id": "my_external_stage"})

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            return_value=mock_response,
        ) as pmock:
            run_test("heif", newtest, includes)

        self.check_mocks_called(pmock)

    def test_global_stage(self, fulltest, mockargs, includes, fake_stages):
        """Load stage from global config"""
        mock_response = Mock(**mockargs)

        stage_includes = []

        newtest = deepcopy(fulltest)
        newtest["includes"] = stage_includes
        newtest["stages"].insert(0, {"type": "ref", "id": "my_external_stage"})

        includes = dataclasses.replace(includes, stages=fake_stages)

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            return_value=mock_response,
        ) as pmock:
            run_test("heif", newtest, includes)

        self.check_mocks_called(pmock)

    def test_both_stages(self, fulltest, mockargs, includes, fake_stages):
        """Load stage defined in both - raise a warning for now"""
        mock_response = Mock(**mockargs)

        stage_includes = [{"stages": fake_stages}]

        newtest = deepcopy(fulltest)
        newtest["includes"] = stage_includes
        newtest["stages"].insert(0, {"type": "ref", "id": "my_external_stage"})

        includes = dataclasses.replace(includes, stages=fake_stages)

        with pytest.raises(exceptions.DuplicateStageDefinitionError):
            with patch(
                "tavern._plugins.rest.request.requests.Session.request",
                return_value=mock_response,
            ) as pmock:
                run_test("heif", newtest, includes)

        assert not pmock.called

    def test_neither(self, fulltest, mockargs, includes, fake_stages):
        """Raises error if not defined"""
        mock_response = Mock(**mockargs)

        stage_includes = []

        newtest = deepcopy(fulltest)
        newtest["includes"] = stage_includes
        newtest["stages"].insert(0, {"type": "ref", "id": "my_external_stage"})

        with pytest.raises(exceptions.InvalidStageReferenceError):
            with patch(
                "tavern._plugins.rest.request.requests.Session.request",
                return_value=mock_response,
            ):
                run_test("heif", newtest, includes)


class TestRetry:
    def test_repeats_twice_and_succeeds(self, fulltest, mockargs, includes):
        fulltest["stages"][0]["max_retries"] = 1
        failed_mockargs = deepcopy(mockargs)
        failed_mockargs["status_code"] = 400

        mock_responses = [Mock(**failed_mockargs), Mock(**mockargs)]

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            side_effect=mock_responses,
        ) as pmock:
            run_test("heif", fulltest, includes)

        assert pmock.call_count == 2

    def test_repeats_twice_and_fails(self, fulltest, mockargs, includes):
        fulltest["stages"][0]["max_retries"] = 1
        mockargs["status_code"] = 400
        mock_response = Mock(**mockargs)

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            return_value=mock_response,
        ) as pmock:
            with pytest.raises(exceptions.TestFailError):
                run_test("heif", fulltest, includes)

        assert pmock.call_count == 2

    def test_run_once(self, fulltest, mockargs, includes):
        mock_responses = Mock(**mockargs)

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            return_value=mock_responses,
        ) as pmock:
            run_test("heif", fulltest, includes)

        assert pmock.call_count == 1


class TestDelay:
    def test_sleep_before(self, fulltest, mockargs, includes):
        """Should sleep with delay_before in stage spec"""

        fulltest["stages"][0]["delay_before"] = 2

        mock_response = Mock(**mockargs)

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            return_value=mock_response,
        ) as pmock:
            with patch("tavern._core.testhelpers.time.sleep") as smock:
                run_test("heif", fulltest, includes)

        assert pmock.called
        smock.assert_called_with(2)

    def test_sleep_after(self, fulltest, mockargs, includes):
        """Should sleep with delay_after in stage spec"""

        fulltest["stages"][0]["delay_after"] = 2

        mock_response = Mock(**mockargs)

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            return_value=mock_response,
        ) as pmock:
            with patch("tavern._core.testhelpers.time.sleep") as smock:
                run_test("heif", fulltest, includes)

        assert pmock.called
        smock.assert_called_with(2)


class TestTavernMetaFormat:
    def test_format_env_keys(self, fulltest, mockargs, includes):
        """Should be able to get variables from the environment and use them in
        test responses"""

        env_key = "SPECIAL_CI_MAGIC_COMMIT_TAG"

        fulltest["stages"][0]["request"]["params"] = {
            "a_format_key": "{tavern.env_vars.%s}" % env_key
        }

        mock_response = Mock(**mockargs)

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            return_value=mock_response,
        ) as pmock:
            with patch.dict(os.environ, {env_key: "bleuihg"}):
                run_test("heif", fulltest, includes)

        assert pmock.called

    def test_format_env_keys_missing_failure(self, fulltest, mockargs, includes):
        """Fails if key is not present"""

        env_key = "SPECIAL_CI_MAGIC_COMMIT_TAG"

        fulltest["stages"][0]["request"]["params"] = {
            "a_format_key": "{tavern.env_vars.%s}" % env_key
        }

        with pytest.raises(exceptions.MissingFormatError):
            run_test("heif", fulltest, includes)


class TestFormatRequestVars:
    @pytest.mark.parametrize("request_key", ("params", "json", "headers"))
    def test_format_request_var_dict(self, fulltest, mockargs, includes, request_key):
        """Variables from request should be available to format in response"""

        sent_value = str(uuid.uuid4())

        fulltest["stages"][0]["request"]["method"] = "POST"
        fulltest["stages"][0]["request"][request_key] = {"a_format_key": sent_value}

        if request_key == "json":
            resp_key = "json"
            mockargs[request_key] = lambda: {"returned": sent_value}
        else:
            resp_key = request_key
            mockargs[request_key] = {"returned": sent_value}

        fulltest["stages"][0]["response"][resp_key] = {
            "returned": "{tavern.request_vars.%s.a_format_key:s}" % request_key
        }

        mock_response = Mock(**mockargs)

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            return_value=mock_response,
        ) as pmock:
            run_test("heif", fulltest, includes)

        assert pmock.called

    @pytest.mark.parametrize("request_key", ("url", "method"))
    def test_format_request_var_value(self, fulltest, mockargs, includes, request_key):
        """Variables from request should be available to format in response"""

        sent_value = str(uuid.uuid4())

        fulltest["stages"][0]["request"]["method"] = "POST"
        fulltest["stages"][0]["request"][request_key] = sent_value

        resp_key = request_key
        mockargs[request_key] = {"returned": sent_value}

        fulltest["stages"][0]["response"][resp_key] = {
            "returned": "{tavern.request_vars.%s:s}" % request_key
        }

        mock_response = Mock(**mockargs)

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            return_value=mock_response,
        ) as pmock, patch(
            "tavern._plugins.rest.request.valid_http_methods", ["POST", sent_value]
        ):
            run_test("heif", fulltest, includes)

        assert pmock.called


class TestFormatMQTTVarsJson:
    """Test that formatting request vars from mqtt works as well, with json payload"""

    @pytest.fixture(name="fulltest")
    def fix_mqtt_publish_test(self):
        spec = {
            "test_name": "An mqtt test with a single stage",
            "paho-mqtt": {
                "connect": {"host": "localhost"},
            },
            "stages": [
                {
                    "name": "step 1",
                    "mqtt_publish": {
                        "topic": "/abc/123",
                        "json": {"message": str(uuid.uuid4())},
                    },
                    "mqtt_response": {
                        "topic": "{tavern.request_vars.topic}",
                        "json": {"echo": "{tavern.request_vars.json.message}"},
                    },
                }
            ],
        }

        return spec

    def test_format_request_var_dict(self, fulltest, includes):
        """Variables from request should be available to format in response -
        this is the original keys in the input file, NOT the formatted ones
        where 'json' is converted to 'payload' in the actual MQTT publish"""

        stage = fulltest["stages"][0]
        sent = stage["mqtt_publish"]["json"]

        mockargs = {
            "spec": paho.MQTTMessage,
            "payload": json.dumps({"echo": sent["message"]}).encode("utf8"),
            "topic": stage["mqtt_publish"]["topic"],
            "timestamp": 0,
        }
        mock_response = Mock(**mockargs)

        fake_client = MagicMock(
            spec=MQTTClient,
            message_received=Mock(return_value=mock_response),
        )

        with patch(
            "tavern._core.run.get_extra_sessions",
            return_value={"paho-mqtt": fake_client},
        ) as pmock:
            run_test("heif", fulltest, includes)

        assert pmock.called


class TestFormatMQTTVarsPlain:
    """Test that formatting request vars from mqtt works as well, with normal payload"""

    @pytest.fixture(name="fulltest")
    def fix_mqtt_publish_test(self):
        spec = {
            "test_name": "An mqtt test with a single stage",
            "paho-mqtt": {
                "connect": {"host": "localhost"},
            },
            "stages": [
                {
                    "name": "step 1",
                    "mqtt_publish": {"topic": "/abc/123", "payload": "hello"},
                    "mqtt_response": {
                        "topic": "{tavern.request_vars.topic}",
                        "payload": "{tavern.request_vars.payload}",
                    },
                }
            ],
        }

        return spec

    def test_format_request_var_value(self, fulltest, includes):
        """Same as above but with plain keys"""
        stage = fulltest["stages"][0]
        sent = stage["mqtt_publish"]["payload"]

        mockargs = {
            "spec": paho.MQTTMessage,
            "payload": sent.encode("utf8"),
            "topic": stage["mqtt_publish"]["topic"],
            "timestamp": 0,
        }
        mock_response = Mock(**mockargs)

        fake_client = MagicMock(
            spec=MQTTClient, message_received=Mock(return_value=mock_response)
        )

        with patch(
            "tavern._core.run.get_extra_sessions",
            return_value={"paho-mqtt": fake_client},
        ) as pmock:
            run_test("heif", fulltest, includes)

        assert pmock.called


class TestFinally:
    @staticmethod
    def run_test(fulltest, mockargs, includes):
        mock_response = Mock(**mockargs)

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            return_value=mock_response,
        ) as pmock:
            run_test("heif", fulltest, includes)

        assert pmock.called

        return pmock

    @pytest.mark.parametrize("finally_block", ([],))
    def test_nop(self, fulltest, mockargs, includes, finally_block):
        """ignore empty finally blocks"""
        fulltest["finally"] = finally_block

        self.run_test(fulltest, mockargs, includes)

    @pytest.mark.parametrize(
        "finally_block",
        (
            {},
            "hi",
            3,
        ),
    )
    def test_wrong_type(self, fulltest, mockargs, includes, finally_block):
        """final stages need to be dicts too"""
        fulltest["finally"] = finally_block

        with pytest.raises(exceptions.BadSchemaError):
            self.run_test(fulltest, mockargs, includes)

    @pytest.fixture
    def finally_request(self):
        return {
            "name": "step 1",
            "request": {"url": "http://www.myfinal.com", "method": "POST"},
            "response": {
                "status_code": 200,
                "json": {"key": "value"},
                "headers": {"content-type": "application/json"},
            },
        }

    def test_finally_run(self, fulltest, mockargs, includes, finally_request):
        fulltest["finally"] = [finally_request]

        pmock = self.run_test(fulltest, mockargs, includes)

        assert pmock.call_count == 2
        assert pmock.mock_calls[1].kwargs.items() >= finally_request["request"].items()

    def test_finally_run_twice(self, fulltest, mockargs, includes, finally_request):
        fulltest["finally"] = [finally_request, finally_request]

        pmock = self.run_test(fulltest, mockargs, includes)

        assert pmock.call_count == 3
        assert pmock.mock_calls[1].kwargs.items() >= finally_request["request"].items()
        assert pmock.mock_calls[2].kwargs.items() >= finally_request["request"].items()

    def test_finally_run_on_main_failure(
        self, fulltest, mockargs, includes, finally_request
    ):
        fulltest["finally"] = [finally_request]

        mockargs["status_code"] = 503

        mock_response = Mock(**mockargs)

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            return_value=mock_response,
        ) as pmock:
            with pytest.raises(exceptions.TestFailError):
                run_test("heif", fulltest, includes)

        assert pmock.call_count == 2
        assert pmock.mock_calls[1].kwargs.items() >= finally_request["request"].items()


class TestTinctures:
    @pytest.mark.parametrize(
        "tinctures",
        (
            {"function": "abc"},
            [{"function": "abc"}],
            [{"function": "abc"}, {"function": "def"}],
        ),
    )
    @pytest.mark.parametrize(
        "at_stage_level",
        (
            True,
            False,
        ),
    )
    def test_tinctures(
        self,
        fulltest,
        mockargs,
        includes,
        tinctures,
        at_stage_level,
    ):
        if at_stage_level:
            fulltest["tinctures"] = tinctures
        else:
            fulltest["stages"][0]["tinctures"] = tinctures

        mock_response = Mock(**mockargs)

        tincture_func_mock = Mock()

        with patch(
            "tavern._plugins.rest.request.requests.Session.request",
            return_value=mock_response,
        ) as pmock, patch(
            "tavern._core.tincture.get_wrapped_response_function",
            return_value=tincture_func_mock,
        ):
            run_test("heif", fulltest, includes)

        assert pmock.call_count == 1
        assert tincture_func_mock.call_count == len(tinctures)


def test_copy_config(pytestconfig):
    cfg_1 = load_global_cfg(pytestconfig)

    cfg_1.variables["test1"] = "abc"

    cfg_2 = load_global_cfg(pytestconfig)

    assert cfg_2.variables.get("test1") is None



from unittest.mock import patch

import pytest

from tavern._core.tincture import Tinctures, get_stage_tinctures


@pytest.fixture(name="example")
def example():
    spec = {
        "test_name": "A test with a single stage",
        "stages": [
            {
                "name": "step 1",
                "request": {"url": "http://www.google.com", "method": "GET"},
                "response": {
                    "status_code": 200,
                    "json": {"key": "value"},
                    "headers": {"content-type": "application/json"},
                },
            }
        ],
    }

    return spec


def test_empty():
    t = Tinctures([])

    t.start_tinctures({})
    t.end_tinctures({}, None)


@pytest.mark.parametrize(
    "tinctures",
    (
        {"function": "abc"},
        [{"function": "abc"}],
        [{"function": "abc"}, {"function": "def"}],
    ),
)
class TestTinctures:
    class TestTinctureBasic:
        def test_stage_tinctures_normal(self, example, tinctures):
            stage = example["stages"][0]
            stage["tinctures"] = tinctures

            with patch(
                "tavern._core.tincture.get_wrapped_response_function",
                return_value=lambda _: None,
            ) as call_mock:
                t = get_stage_tinctures(stage, example)

            t.start_tinctures(stage)
            t.end_tinctures(stage, None)

            assert call_mock.call_count == len(tinctures)

        def test_test_tinctures_normal(self, example, tinctures):
            stage = example["stages"][0]
            example["tinctures"] = tinctures

            with patch(
                "tavern._core.tincture.get_wrapped_response_function",
                return_value=lambda _: None,
            ) as call_mock:
                t = get_stage_tinctures(stage, example)

            t.start_tinctures(stage)
            t.end_tinctures(stage, None)

            assert call_mock.call_count == len(tinctures)

    class TestTinctureYields:
        @staticmethod
        def does_yield(stage):
            assert stage["name"] == "step 1"

            (expected, response) = yield

            assert expected == stage["response"]
            assert response is None

        def test_stage_tinctures_normal(self, example, tinctures):
            stage = example["stages"][0]
            stage["tinctures"] = tinctures

            with patch(
                "tavern._core.tincture.get_wrapped_response_function",
                return_value=self.does_yield,
            ) as call_mock:
                t = get_stage_tinctures(stage, example)

            t.start_tinctures(stage)
            t.end_tinctures(stage["response"], None)

            assert call_mock.call_count == len(tinctures)

        def test_test_tinctures_normal(self, example, tinctures):
            stage = example["stages"][0]
            example["tinctures"] = tinctures

            with patch(
                "tavern._core.tincture.get_wrapped_response_function",
                return_value=self.does_yield,
            ) as call_mock:
                t = get_stage_tinctures(stage, example)

            t.start_tinctures(stage)
            t.end_tinctures(stage["response"], None)

            assert call_mock.call_count == len(tinctures)



import logging
from os.path import abspath, dirname, join
from typing import Dict, Iterable, Optional, Union

import yaml

from tavern._core.dict_util import format_keys
from tavern._core.pytest.config import TestConfig

from .client import MQTTClient
from .request import MQTTRequest
from .response import MQTTResponse

logger: logging.Logger = logging.getLogger(__name__)

session_type = MQTTClient

request_type = MQTTRequest
request_block_name = "mqtt_publish"


def get_expected_from_request(
    response_block: Union[Dict, Iterable[Dict]],
    test_block_config: TestConfig,
    session: MQTTClient,
) -> Optional[Dict]:
    expected: Optional[Dict] = None

    # mqtt response is not required
    if response_block:
        expected = {"mqtt_responses": []}
        if isinstance(response_block, dict):
            response_block = [response_block]

        for response in response_block:
            # format so we can subscribe to the right topic
            f_expected = format_keys(response, test_block_config.variables)
            mqtt_client = session
            mqtt_client.subscribe(f_expected["topic"], f_expected.get("qos", 1))
            expected["mqtt_responses"].append(f_expected)

    return expected


verifier_type = MQTTResponse
response_block_name = "mqtt_response"

schema_path: str = join(abspath(dirname(__file__)), "jsonschema.yaml")
with open(schema_path, encoding="utf-8") as schema_file:
    schema = yaml.load(schema_file, Loader=yaml.SafeLoader)



import datetime

import jwt


def assert_quick_response(response):
    """Make sure that a request doesn't take too long

    Args:
        response (requests.Response): response object
    """
    assert response.elapsed < datetime.timedelta(seconds=0.1)


def create_bearer_token():
    # Authorization: "bearer {test_login_token:s}"

    SECRET = "CGQgaG7GYvTcpaQZqosLy4"
    SERVERNAME = "testserver"

    payload = {
        "sub": "test-user",
        "aud": SERVERNAME,
        "exp": datetime.datetime.utcnow() + datetime.timedelta(hours=1),
    }

    token = jwt.encode(payload, SECRET, algorithm="HS256")

    return {"Authorization": "Bearer {}".format(token)}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: helloworld_v1_precompiled.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1fhelloworld_v1_precompiled.proto\x12\rhelloworld.v1"\x1c\n\x0cHelloRequest\x12\x0c\n\x04name\x18\x01 \x01(\t"\x1d\n\nHelloReply\x12\x0f\n\x07message\x18\x01 \x01(\t2O\n\x07Greeter\x12\x44\n\x08SayHello\x12\x1b.helloworld.v1.HelloRequest\x1a\x19.helloworld.v1.HelloReply"\x00\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "helloworld_v1_precompiled_pb2", _globals
)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_HELLOREQUEST"]._serialized_start = 50
    _globals["_HELLOREQUEST"]._serialized_end = 78
    _globals["_HELLOREPLY"]._serialized_start = 80
    _globals["_HELLOREPLY"]._serialized_end = 109
    _globals["_GREETER"]._serialized_start = 111
    _globals["_GREETER"]._serialized_end = 190
# @@protoc_insertion_point(module_scope)



import logging
import logging.config
import os
import pathlib
import re
import typing
from textwrap import dedent
from typing import Optional

if typing.TYPE_CHECKING:
    from .file import YamlFile


import pytest
import yaml

from tavern._core import exceptions

from .util import add_ini_options, add_parser_options, get_option_generic


def pytest_addoption(parser: pytest.Parser) -> None:
    add_parser_options(parser.addoption, with_defaults=False)
    add_ini_options(parser)


def pytest_collect_file(parent, path: os.PathLike) -> Optional["YamlFile"]:
    """On collecting files, get any files that end in .tavern.yaml or .tavern.yml as tavern
    test files
    """

    if int(pytest.__version__.split(".", maxsplit=1)[0]) < 7:
        raise exceptions.TavernException("Only pytest >=7 is supported")

    try:
        setup_initial_logging = get_option_generic(
            parent.config, "tavern-setup-init-logging", False
        )
    except ValueError:
        pass
    else:
        if setup_initial_logging:
            cfg = dedent(
                """
            ---
            version: 1
            formatters:
              default:
                format: "%(asctime)s [%(levelname)s]: (%(name)s:%(lineno)d) %(message)s"
                style: "%"
                datefmt: "%X"

            handlers:
              stderr:
                class : logging.StreamHandler
                level   : DEBUG
                formatter: default
                stream  : ext://sys.stderr

            loggers:
              tavern:
                handlers:
                  - stderr
                level: DEBUG
            """
            )

            settings = yaml.load(cfg, Loader=yaml.SafeLoader)
            logging.config.dictConfig(settings)

    pattern = get_option_generic(
        parent.config, "tavern-file-path-regex", r".+\.tavern\.ya?ml$"
    )

    if isinstance(pattern, list):
        if len(pattern) != 1:
            raise exceptions.InvalidConfigurationException(
                "tavern-file-path-regex must have exactly one option"
            )
        pattern = pattern[0]

    try:
        compiled = re.compile(pattern)
    except Exception as e:
        raise exceptions.InvalidConfigurationException(e) from e

    match_tavern_file = compiled.search

    from .file import YamlFile

    path = pathlib.Path(path)

    if match_tavern_file(str(path)):
        return YamlFile.from_parent(parent, path=path)

    return None


def pytest_addhooks(pluginmanager) -> None:
    """Add our custom tavern hooks"""
    from . import newhooks

    pluginmanager.add_hookspecs(newhooks)



import os
from contextlib import ExitStack
from typing import Union

import pytest
from _pytest.config import ExitCode

from tavern._core import exceptions
from tavern._core.schema.files import wrapfile


def _get_or_wrap_global_cfg(
    stack: ExitStack, tavern_global_cfg: Union[dict, str]
) -> str:
    """
    Try to parse global configuration from given argument.

    Args:
        stack: context stack for wrapping file if a dictionary is given
        tavern_global_cfg: path to a file or a dictionary with configuration.

    Returns:
        path to global config file

    Raises:
        InvalidSettingsError: If global config was not of the right type or a given path
            does not exist
    """

    if isinstance(tavern_global_cfg, str):
        if not os.path.exists(tavern_global_cfg):
            raise exceptions.InvalidSettingsError(
                f"global config file '{tavern_global_cfg}' does not exist"
            )
        global_filename = tavern_global_cfg
    elif isinstance(tavern_global_cfg, dict):
        global_filename = stack.enter_context(wrapfile(tavern_global_cfg))
    else:
        raise exceptions.InvalidSettingsError(
            "Invalid format for global settings - must be dict or path to settings file, was {}".format(
                type(tavern_global_cfg)
            )
        )

    return global_filename


def run(  # type:ignore
    in_file: str,
    tavern_global_cfg: Union[dict, str, None] = None,
    tavern_mqtt_backend: Union[str, None] = None,
    tavern_http_backend: Union[str, None] = None,
    tavern_grpc_backend: Union[str, None] = None,
    tavern_strict: Union[bool, None] = None,
    pytest_args: Union[list, None] = None,
) -> Union[ExitCode, int]:
    """Run all tests contained in a file using pytest.main()

    Args:
        in_file: file to run tests on
        tavern_global_cfg: Extra global config
        tavern_mqtt_backend: name of MQTT plugin to use. If not
            specified, uses tavern-mqtt
        tavern_http_backend: name of HTTP plugin to use. If not
            specified, use tavern-http
        tavern_grpc_backend: name of GRPC plugin to use. If not
            specified, use tavern-grpc
        tavern_strict: Strictness of checking for responses.
            See documentation for details
        pytest_args: List of extra arguments to pass directly
            to Pytest as if they were command line arguments

    Returns:
        Whether ALL tests passed or not
    """

    pytest_args = pytest_args or []
    pytest_args += [in_file]

    if tavern_mqtt_backend:
        pytest_args += ["--tavern-mqtt-backend", tavern_mqtt_backend]
    if tavern_http_backend:
        pytest_args += ["--tavern-http-backend", tavern_http_backend]
    if tavern_grpc_backend:
        pytest_args += ["--tavern-grpc-backend", tavern_grpc_backend]
    if tavern_strict:
        pytest_args += ["--tavern-strict", tavern_strict]

    with ExitStack() as stack:
        if tavern_global_cfg:
            global_filename = _get_or_wrap_global_cfg(stack, tavern_global_cfg)
            pytest_args += ["--tavern-global-cfg", global_filename]
        return pytest.main(args=pytest_args)



import os.path
import shutil
import tempfile

import pytest


@pytest.fixture()
def make_temp_dir():
    with tempfile.TemporaryDirectory() as d:
        yield d


@pytest.fixture(autouse=True, scope="session")
def single_compiled_proto_for_test():
    with tempfile.TemporaryDirectory() as d:
        proto_filename = "helloworld_v2_compiled.proto"
        dst = os.path.join(d, proto_filename)
        shutil.copy(proto_filename, dst)
        yield dst



import logging
import os
from typing import List, Union

from tavern._core.loader import load_single_document_yaml

from .dict_util import deep_dict_merge

logger: logging.Logger = logging.getLogger(__name__)


def load_global_config(global_cfg_paths: List[Union[str, os.PathLike]]) -> dict:
    """Given a list of file paths to global config files, load each of them and
    return the joined dictionary.

    This does a deep dict merge.

    Args:
        global_cfg_paths: List of filenames to load from

    Returns:
        joined global configs
    """
    global_cfg: dict = {}

    if global_cfg_paths:
        logger.debug("Loading global config from %s", global_cfg_paths)
        for filename in global_cfg_paths:
            contents = load_single_document_yaml(filename)
            global_cfg = deep_dict_merge(global_cfg, contents)

    return global_cfg


valid_http_methods = [
    "GET",
    "PUT",
    "POST",
    "DELETE",
    "PATCH",
    "OPTIONS",
    "HEAD",
]



import collections.abc
import dataclasses
import inspect
import logging
from typing import Any, Generator, List

from tavern._core import exceptions
from tavern._core.extfunctions import get_wrapped_response_function

logger: logging.Logger = logging.getLogger(__name__)


@dataclasses.dataclass
class Tinctures:
    tinctures: List[Any]
    needs_response: List[Generator] = dataclasses.field(default_factory=list)

    def start_tinctures(self, stage: collections.abc.Mapping) -> None:
        results = [t(stage) for t in self.tinctures]

        for r in results:
            if inspect.isgenerator(r):
                # Store generator and start it
                self.needs_response.append(r)
                next(r)

    def end_tinctures(self, expected: collections.abc.Mapping, response) -> None:
        """
        Send the response object to any tinctures that want it

        Args:
            expected: 'expected' from initial test - type varies depending on backend
            response: The response from 'run' for the stage - type varies depending on backend
        """
        if self.needs_response is None:
            raise RuntimeError(
                "should not be called before accumulating tinctures which need a response"
            )

        for n in self.needs_response:
            try:
                n.send((expected, response))
            except StopIteration:
                pass
            else:
                raise RuntimeError("Tincture had more than one yield")


def get_stage_tinctures(
    stage: collections.abc.Mapping, test_spec: collections.abc.Mapping
) -> Tinctures:
    """Get tinctures for stage

    Args:
        stage: Stage
        test_spec: Whole test spec
    """
    stage_tinctures = []

    def add_tinctures_from_block(maybe_tinctures, blockname: str):
        logger.debug("Trying to add tinctures from %s", blockname)

        def inner_yield():
            if maybe_tinctures is not None:
                if isinstance(maybe_tinctures, list):
                    for vf in maybe_tinctures:
                        yield get_wrapped_response_function(vf)
                elif isinstance(maybe_tinctures, dict):
                    yield get_wrapped_response_function(maybe_tinctures)
                elif maybe_tinctures is not None:
                    raise exceptions.BadSchemaError(
                        f"Badly formatted 'tinctures' block in {blockname}"
                    )

        stage_tinctures.extend(inner_yield())

    add_tinctures_from_block(test_spec.get("tinctures"), "test")
    add_tinctures_from_block(stage.get("tinctures"), "stage")

    logger.debug("%d tinctures for stage %s", len(stage_tinctures), stage["name"])

    return Tinctures(stage_tinctures)



import time


def return_hello():
    return {"hello": "there"}


def return_goodbye_string():
    return "goodbye"


def return_list_vals():
    return [{"a_value": "b_value"}, 2]


def gen_echo_url(host):
    return f"{host}/echo"


def time_request(_):
    time.time()
    yield
    time.time()


def print_response(_, extra_print="affa"):
    (_, r) = yield



import pytest


@pytest.fixture
def str_fixture():
    return "abc-fixture-value"


@pytest.fixture(name="yield_str_fixture")
def sdkofsok(str_fixture):
    yield str_fixture


@pytest.fixture(name="yielder")
def bluerhug(request):
    # This doesn't really do anything at the moment. In future it might yield
    # the result or something, but it's a bit difficult to do at the moment.
    yield "hello"


@pytest.fixture(scope="session", autouse=True)
def autouse_thing():
    return "abc"


@pytest.fixture(scope="session", autouse=True)
def fixture_echo_url():
    return "http://localhost:5003/echo"


@pytest.fixture(scope="session", autouse=True, name="autouse_thing_named")
def second(autouse_thing):
    return autouse_thing



import concurrent
import concurrent.futures
import contextlib
import itertools
import json
import logging
import time
from dataclasses import dataclass
from typing import Dict, List, Mapping, Optional, Tuple, Union

from paho.mqtt.client import MQTTMessage

from tavern._core import exceptions
from tavern._core.dict_util import check_keys_match_recursive
from tavern._core.loader import ANYTHING
from tavern._core.pytest.config import TestConfig
from tavern._core.pytest.newhooks import call_hook
from tavern._core.report import attach_yaml
from tavern._core.strict_util import StrictOption
from tavern.response import BaseResponse

from .client import MQTTClient

logger: logging.Logger = logging.getLogger(__name__)

_default_timeout = 1


class MQTTResponse(BaseResponse):
    response: MQTTMessage

    def __init__(
        self,
        client: MQTTClient,
        name: str,
        expected: TestConfig,
        test_block_config: TestConfig,
    ) -> None:
        super().__init__(name, expected, test_block_config)

        self._client = client

        self.received_messages: List = []

    def __str__(self) -> str:
        if self.response:
            return self.response.payload.decode("utf-8")
        else:
            return "<Not run yet>"

    def verify(self, response: MQTTMessage) -> Mapping:
        """Ensure mqtt message has arrived

        Args:
            response: not used except for debug printing
        """

        self.response = response

        try:
            return self._await_response()
        finally:
            self._client.unsubscribe_all()

    def _await_response(self) -> Mapping:
        """Actually wait for response

        Returns:
            things to save to variables for the rest of this test
        """

        # Get into class with metadata attached
        expected = self.expected["mqtt_responses"]

        by_topic = {
            m: list(v) for m, v in itertools.groupby(expected, lambda x: x["topic"])
        }

        correct_messages: List["_ReturnedMessage"] = []
        warnings: List[str] = []

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = []

            for topic, expected_for_topic in by_topic.items():
                logger.debug("Starting thread for messages on topic '%s'", topic)
                futures.append(
                    executor.submit(
                        self._await_messages_on_topic, topic, expected_for_topic
                    )
                )

            for future in concurrent.futures.as_completed(futures):
                # for future in futures:
                try:
                    messages, warnings = future.result()
                except Exception as e:
                    raise exceptions.ConcurrentError(
                        "unexpected error getting result from future"
                    ) from e
                else:
                    warnings.extend(warnings)
                    correct_messages.extend(messages)

        if self.errors:
            if warnings:
                self._adderr("\n".join(warnings))

            raise exceptions.TestFailError(
                f"Test '{self.name:s}' failed:\n{self._str_errors():s}",
                failures=self.errors,
            )

        saved: Dict = {}

        for msg in correct_messages:
            # Check saving things from the payload and from json
            saved.update(
                self.maybe_get_save_values_from_save_block(
                    "payload",
                    msg.msg.payload,
                    outer_save_block=msg.expected,
                )
            )
            saved.update(
                self.maybe_get_save_values_from_save_block(
                    "json",
                    msg.msg.payload,
                    outer_save_block=msg.expected,
                )
            )

            saved.update(self.maybe_get_save_values_from_ext(msg.msg, msg.expected))

        # Trying to save might have introduced errors, so check again
        if self.errors:
            raise exceptions.TestFailError(
                "Saving results from test '{:s}' failed:\n{:s}".format(
                    self.name, self._str_errors()
                ),
                failures=self.errors,
            )

        return saved

    def _await_messages_on_topic(
        self, topic: str, expected: List[Dict]
    ) -> Tuple[List["_ReturnedMessage"], List[str]]:
        """
        Waits for the specific message

        Args:
            topic: topic to listen on
            expected: expected response for this block

        Returns:
            The correct message (if any) and warnings from processing the message
        """

        timeout = max(m.get("timeout", _default_timeout) for m in expected)

        # A list of verifiers that can be used to validate messages for this topic
        verifiers = [_MessageVerifier(self.test_block_config, v) for v in expected]

        correct_messages = []
        warnings = []

        time_spent = 0.0
        while (time_spent < timeout) and verifiers:
            t0 = time.time()

            msg = self._client.message_received(topic, timeout - time_spent)

            if not msg:
                # timed out
                break

            logger.debug("Seeing if message '%s' matched expected", msg)

            call_hook(
                self.test_block_config,
                "pytest_tavern_beta_after_every_response",
                expected=expected,
                response=msg,
            )

            self.received_messages.append(msg)

            with contextlib.suppress(AttributeError):
                msg.payload = msg.payload.decode("utf8")

            attach_yaml(
                {
                    "topic": msg.topic,
                    "payload": msg.payload,
                    "timestamp": msg.timestamp,
                },
                name="rest_response",
            )

            found: List[int] = []
            for i, v in enumerate(verifiers):
                if v.is_valid(msg):
                    correct_messages.append(_ReturnedMessage(v.expected, msg))
                    if found:
                        logger.warning(
                            "Message was matched by multiple mqtt_response blocks"
                        )
                    found.append(i)
                warnings.extend(v.popwarnings())
            verifiers = [v for (i, v) in enumerate(verifiers) if i not in found]

            time_spent += time.time() - t0

        if verifiers:
            for v in verifiers:
                self._adderr(
                    "Expected '%s' on topic '%s' but no such message received",
                    v.expected_payload,
                    topic,
                )

        for msg in correct_messages:
            if msg.expected.get("unexpected"):
                self._adderr(
                    "Got '%s' on topic '%s' marked as unexpected",
                    msg.expected["payload"],
                    topic,
                )

            self._maybe_run_validate_functions(msg)

        return correct_messages, warnings


@dataclass
class _ReturnedMessage:
    """An actual message returned from the API and it's matching 'expected' block."""

    expected: Mapping
    msg: MQTTMessage


class _MessageVerifier:
    def __init__(self, test_block_config: TestConfig, expected: Mapping) -> None:
        self.expires = time.time() + expected.get("timeout", _default_timeout)

        self.expected = expected
        self.expected_payload, self.expect_json_payload = self._get_payload_vals(
            expected
        )

        test_strictness = test_block_config.strict
        self.block_strictness: StrictOption = test_strictness.option_for("json")

        # Any warnings to do with the request
        # eg, if a message was received but it didn't match, message had payload, etc.
        self.warnings: List[str] = []

    def is_valid(self, msg: MQTTMessage) -> bool:
        if time.time() > self.expires:
            return False

        topic = self.expected["topic"]

        def addwarning(w, *args, **kwargs):
            logger.warning(w, *args, **kwargs)
            self.warnings.append(w % args)

        if self.expect_json_payload:
            try:
                msg.payload = json.loads(msg.payload)
            except json.decoder.JSONDecodeError:
                addwarning(
                    "Expected a json payload but got '%s'",
                    msg.payload,
                    exc_info=True,
                )
                return False

        if self.expected_payload is None:
            if msg.payload is None or msg.payload == "":
                logger.info("Got message with no payload (as expected) on '%s'", topic)
                return True
            else:
                addwarning(
                    "Message had payload '%s' but we expected no payload",
                    msg.payload,
                )
        elif self.expected_payload is ANYTHING:
            logger.info("Got message on %s matching !anything token", topic)
            return True
        elif msg.payload != self.expected_payload:
            if self.expect_json_payload:
                try:
                    check_keys_match_recursive(
                        self.expected_payload,
                        msg.payload,
                        [],
                        strict=self.block_strictness,
                    )
                except exceptions.KeyMismatchError:
                    # Just want to log the mismatch
                    pass
                else:
                    logger.info(
                        "Got expected message in '%s' with expected payload",
                        msg.topic,
                    )
                    logger.debug("Matched payload was '%s", msg.payload)
                    return True

            addwarning(
                "Got unexpected payload on topic '%s': '%s' (expected '%s')",
                msg.topic,
                msg.payload,
                self.expected_payload,
            )
        else:
            logger.info(
                "Got expected message in '%s' with expected payload",
                msg.topic,
            )
            logger.debug("Matched payload was '%s", msg.payload)
            return True

        return False

    @staticmethod
    def _get_payload_vals(expected: Mapping) -> Tuple[Optional[Union[str, dict]], bool]:
        """Gets the payload from the 'expected' block

        Returns:
            First element is the expected payload, second element is whether it's
                expected to be json or not
        """
        # TODO move this check to initialisation/schema checking
        if "json" in expected:
            if "payload" in expected:
                raise exceptions.BadSchemaError(
                    "Can only specify one of 'payload' or 'json' in MQTT response"
                )

            payload = expected["json"]
            json_payload = True

            if payload.pop("$ext", None):
                raise exceptions.MisplacedExtBlockException(
                    "json",
                )
        elif "payload" in expected:
            payload = expected["payload"]
            json_payload = False
        else:
            payload = None
            json_payload = False

        return payload, json_payload

    def popwarnings(self) -> List[str]:
        popped = []
        while self.warnings:
            popped.append(self.warnings.pop(0))
        return popped



import argparse
import logging.config
from argparse import ArgumentParser
from textwrap import dedent
from typing import Dict

from .core import run


class TavernArgParser(ArgumentParser):
    def __init__(self) -> None:
        description = """Parse yaml + make requests against an API

        Any extra arguments will be passed directly to Pytest. Run py.test --help for a list"""

        super().__init__(
            description=dedent(description),
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

        self.add_argument("in_file", help="Input file with tests in")

        self.add_argument(
            "--log-to-file",
            help="Log output to a file (tavern.log if no argument is given)",
            nargs="?",
            const="tavern.log",
        )

        self.add_argument(
            "--stdout", help="Log output stdout", action="store_true", default=False
        )

        self.add_argument(
            "--debug",
            help="Log debug information (only relevant if --stdout or --log-to-file is passed)",
            action="store_true",
            default=False,
        )


def main():
    args, remaining = TavernArgParser().parse_known_args()
    vargs = vars(args)

    if vargs.pop("debug"):
        log_level = "DEBUG"
    else:
        log_level = "INFO"

    # Basic logging config that will print out useful information
    log_cfg: Dict = {
        "version": 1,
        "formatters": {
            "default": {
                "format": "%(asctime)s [%(levelname)s]: (%(name)s:%(lineno)d) %(message)s",
                "style": "%",
            }
        },
        "handlers": {
            "to_stdout": {
                "class": "logging.StreamHandler",
                "formatter": "default",
                "stream": "ext://sys.stdout",
            },
            "nothing": {"class": "logging.NullHandler"},
        },
        "loggers": {
            "tavern": {"handlers": ["nothing"], "level": log_level},
            "": {"handlers": ["nothing"], "level": log_level},
        },
    }

    log_loc = vargs.pop("log_to_file")

    if log_loc:
        log_cfg["handlers"].update(
            {
                "to_file": {
                    "class": "logging.FileHandler",
                    "filename": log_loc,
                    "formatter": "default",
                }
            }
        )

        log_cfg["loggers"]["tavern"]["handlers"].append("to_file")

    if vargs.pop("stdout"):
        log_cfg["loggers"]["tavern"]["handlers"].append("to_stdout")

    logging.config.dictConfig(log_cfg)

    in_file = vargs.pop("in_file")
    global_cfg = vargs.pop("tavern_global_cfg", {})

    raise SystemExit(run(in_file, global_cfg, pytest_args=remaining, **vargs))



import os
import re
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Mapping, Tuple, Type, Union

from pykwalify.types import is_bool, is_float, is_int

from tavern._core import exceptions
from tavern._core.exceptions import BadSchemaError
from tavern._core.extfunctions import (
    get_pykwalify_logger,
    import_ext_function,
    is_ext_function,
)
from tavern._core.general import valid_http_methods
from tavern._core.loader import (
    ApproxScalar,
    BoolToken,
    FloatToken,
    IntToken,
    TypeConvertToken,
)
from tavern._core.strict_util import StrictLevel

if TYPE_CHECKING:
    from tavern._plugins.grpc.response import GRPCCode


# To extend pykwalify's type validation, extend its internal functions
# These return boolean values
def validate_type_and_token(
    validate_type: Callable[[Any], bool], token: Type[TypeConvertToken]
):
    def validate(value):
        return validate_type(value) or isinstance(value, token)

    return validate


is_int_like = validate_type_and_token(is_int, IntToken)
is_float_like = validate_type_and_token(is_float, FloatToken)
is_bool_like = validate_type_and_token(is_bool, BoolToken)


# These plug into the pykwalify extension function API
def validator_like(validate: Callable[[Any], bool], description: str):
    def validator(value, rule_obj, path):
        if validate(value):
            return True
        else:
            err_msg = "expected '{}' type at '{}', got '{}'".format(
                description, path, value
            )
            raise BadSchemaError(err_msg)

    return validator


int_variable = validator_like(is_int_like, "int-like")
float_variable = validator_like(is_float_like, "float-like")
bool_variable = validator_like(is_bool_like, "bool-like")


def _validate_one_extension(input_value: Mapping) -> None:
    expected_keys = {"function", "extra_args", "extra_kwargs"}
    extra = set(input_value) - expected_keys

    if extra:
        raise BadSchemaError(f"Unexpected keys passed to $ext: {extra}")

    if "function" not in input_value:
        raise BadSchemaError("No function specified for validation")

    try:
        import_ext_function(input_value["function"])
    except Exception as e:
        raise BadSchemaError("Couldn't load {}".format(input_value["function"])) from e

    extra_args = input_value.get("extra_args")
    extra_kwargs = input_value.get("extra_kwargs")

    if extra_args and not isinstance(extra_args, list):
        raise BadSchemaError(f"Expected a list of extra_args, got {type(extra_args)}")

    if extra_kwargs and not isinstance(extra_kwargs, dict):
        raise BadSchemaError(f"Expected a dict of extra_kwargs, got {type(extra_args)}")


def validate_extensions(value, rule_obj, path) -> bool:
    """Given a specification for calling a validation function, make sure that
    the arguments are valid (ie, function is valid, arguments are of the
    correct type...)

    Arguments/return values are sort of pykwalify internals (this function is
    only called from pykwalify) so not listed

    Todo:
        Because this is loaded by pykwalify as a file, we need some kind of
        entry point to set up logging. Or just fork pykwalify and fix the
        various issues in it.

        We should also check the function signature using the `inspect` module

    Raises:
        BadSchemaError: Something in the validation function spec was wrong
    """

    if isinstance(value, list):
        for vf in value:
            _validate_one_extension(vf)
    elif isinstance(value, dict):
        _validate_one_extension(value)

    return True


def validate_status_code_is_int_or_list_of_ints(value: Mapping, rule_obj, path) -> bool:
    err_msg = "status_code has to be an integer or a list of integers (got {})".format(
        value
    )

    if not isinstance(value, list) and not is_int_like(value):
        raise BadSchemaError(err_msg)

    if isinstance(value, list):
        if not all(is_int_like(i) for i in value):
            raise BadSchemaError(err_msg)

    return True


def check_usefixtures(value: Mapping, rule_obj, path) -> bool:
    err_msg = "'usefixtures' has to be a list with at least one item"

    if not isinstance(value, (list, tuple)):
        raise BadSchemaError(err_msg)

    if not value:
        raise BadSchemaError(err_msg)

    return True


def validate_grpc_status_is_valid_or_list_of_names(
    value: "GRPCCode", rule_obj, path
) -> bool:
    """Validate GRPC statuses https://github.com/grpc/grpc/blob/master/doc/statuscodes.md"""
    # pylint: disable=unused-argument
    err_msg = (
        "status has to be an valid grpc status code, name, or list (got {})".format(
            value
        )
    )

    if isinstance(value, (str, int)):
        if not to_grpc_status(value):
            raise BadSchemaError(err_msg)
    elif isinstance(value, list):
        if not all(to_grpc_status(i) for i in value):
            raise BadSchemaError(err_msg)
    else:
        raise BadSchemaError(err_msg)

    return True


def to_grpc_status(value: Union[str, int]):
    from grpc import StatusCode

    if isinstance(value, str):
        value = value.upper()
        for status in StatusCode:
            if status.name == value:
                return status.name
    elif isinstance(value, int):
        for status in StatusCode:
            if status.value[0] == value:
                return status.name

    return None


def verify_oneof_id_name(value: Mapping, rule_obj, path) -> bool:
    """Checks that if 'name' is not present, 'id' is"""

    if not (name := value.get("name")):
        if name == "":
            raise BadSchemaError("Name cannot be empty")

        if not value.get("id"):
            raise BadSchemaError("If 'name' is not specified, 'id' must be specified")

    return True


def check_parametrize_marks(value, rule_obj, path) -> bool:
    key_or_keys = value["key"]
    vals = value["vals"]

    # At this point we can assume vals is a list - check anyway
    if not (isinstance(vals, list) or is_ext_function(vals)):
        raise BadSchemaError("'vals' should be a list")

    if isinstance(key_or_keys, str):
        # Vals can be anything
        return True
    elif isinstance(key_or_keys, list):
        err_msg = "If 'key' is a list, 'vals' must be a list of lists where each list is the same length as 'key'"

        # Checking for whether the ext function actually returns the correct
        # values has to be deferred until the point where the function is
        # actually called
        if not is_ext_function(vals):
            # broken example:
            # - parametrize:
            #     key:
            #       - edible
            #       - fruit
            #     vals:
            #       a: b
            if not isinstance(vals, list):
                raise BadSchemaError(err_msg)

            # example:
            # - parametrize:
            #     key:
            #       - edible
            #       - fruit
            #     vals:
            #       - [rotten, apple]
            #       - [fresh, orange]
            #       - [unripe, pear]
            for v in vals:
                if not isinstance(v, list):
                    # This catches the case like
                    #
                    # - parametrize:
                    #     key:
                    #       - edible
                    #       - fruit
                    #     vals:
                    #       - fresh
                    #       - orange
                    #
                    # This will parametrize 'edible' as [f, r, e, s, h] which is almost certainly not desired
                    raise BadSchemaError(err_msg)
                if len(v) != len(key_or_keys):
                    # If the 'vals' list has more or less keys
                    raise BadSchemaError(err_msg)

    else:
        raise BadSchemaError("'key' must be a string or a list")

    return True


def validate_data_key(value, rule_obj, path: str) -> bool:
    """Validate the 'data' key in a http request

    From requests docs:

    > data - (optional) Dictionary or list of tuples [(key, value)] (will be
    > form-encoded), bytes, or file-like object to send in the body of the
    > Request.

    We could handle lists of tuples, but it seems entirely pointless to maintain
    compatibility for something which is more verbose and does the same thing
    """

    if isinstance(value, dict):
        # Fine
        pass
    elif isinstance(value, (str, bytes)):
        # Also fine - might want to do checking on this for encoding etc?
        pass
    elif isinstance(value, list):
        raise BadSchemaError(f"Error at {path} - expected a dict, str, or !!binary")

        # invalid = []

        # # Check they're all a list of 2-tuples
        # for p in value:
        #     if not isinstance(p, list):
        #         invalid += p
        #     elif len(p) != 2:
        #         invalid += p

        # if invalid:
        #     raise BadSchemaError("Error at {} - when passing a list to the 'data' key, all items must be 2-tuples (invalid values: {})".format(path, invalid))
    else:
        raise BadSchemaError(f"Error at {path} - expected a dict, str, or !!binary")

    return True


def validate_request_json(value, rule_obj, path) -> bool:
    """Performs the above match, but also matches a dict or a list. This it
    just because it seems like you can't match a dict OR a list in pykwalify
    """

    def nested_values(d):
        if isinstance(d, dict):
            for v in d.values():
                if isinstance(v, dict):
                    yield from v.values()
                else:
                    yield v
        else:
            yield d

    if any(isinstance(i, ApproxScalar) for i in nested_values(value)):
        # If this is a request data block
        if not re.search(r"^/stages/\d/(response/json|mqtt_response/json)", path):
            raise BadSchemaError(
                "Error at {} - Cannot use a '!approx' in anything other than an expected http response body or mqtt response json".format(
                    path
                )
            )

    return True


def validate_json_with_ext(value, rule_obj, path) -> bool:
    """Validate json with extensions"""
    validate_request_json(value, rule_obj, path)

    if isinstance(value, dict):
        maybe_ext_val = value.get("$ext", None)
        if isinstance(maybe_ext_val, dict):
            validate_extensions(maybe_ext_val, rule_obj, path)
        elif maybe_ext_val is not None:
            raise BadSchemaError(f"Unexpected $ext key in block at {path}")

    return True


def check_strict_key(value: Union[List, bool], rule_obj, path) -> bool:
    """Make sure the 'strict' key is either a bool or a list"""

    if not isinstance(value, list) and not is_bool_like(value):
        raise BadSchemaError("'strict' has to be either a boolean or a list")
    elif isinstance(value, list):
        try:
            # Reuse validation here
            StrictLevel.from_options(value)
        except exceptions.InvalidConfigurationException as e:
            raise BadSchemaError from e

    # Might be a bool as well, in which case it's processed further down the line - no validation required

    return True


def validate_timeout_tuple_or_float(value: Union[List, Tuple], rule_obj, path) -> bool:
    """Make sure timeout is a float/int or a tuple of floats/ints"""

    err_msg = "'timeout' must be either a float/int or a 2-tuple of floats/ints - got '{}' (type {})".format(
        value, type(value)
    )
    logger = get_pykwalify_logger("tavern.schemas.extensions")

    def check_is_timeout_val(v):
        if v is True or v is False or not (is_float_like(v) or is_int_like(v)):
            logger.debug("'timeout' value not a float/int")
            raise BadSchemaError(err_msg)

    if isinstance(value, (list, tuple)):
        if len(value) != 2:
            raise BadSchemaError(err_msg)
        for v in value:
            check_is_timeout_val(v)
    else:
        check_is_timeout_val(value)

    return True


def validate_verify_bool_or_str(value: Union[bool, str], rule_obj, path) -> bool:
    """Make sure the 'verify' key is either a bool or a str"""

    if not isinstance(value, (bool, str)) and not is_bool_like(value):
        raise BadSchemaError(
            "'verify' has to be either a boolean or the path to a CA_BUNDLE file or directory with certificates of trusted CAs"
        )

    return True


def validate_cert_tuple_or_str(value, rule_obj, path) -> bool:
    """Make sure the 'cert' key is either a str or tuple"""

    err_msg = (
        "The 'cert' key must be the path to a single file (containing the private key and the certificate) "
        "or as a tuple of both files"
    )

    if not isinstance(value, (str, tuple, list)):
        raise BadSchemaError(err_msg)

    if isinstance(value, (list, tuple)):
        if len(value) != 2:
            raise BadSchemaError(err_msg)
        elif not all(isinstance(i, str) for i in value):
            raise BadSchemaError(err_msg)

    return True


def validate_file_spec(value: Dict, rule_obj, path) -> bool:
    """Validate file upload arguments"""

    logger = get_pykwalify_logger("tavern.schema.extensions")

    if not isinstance(value, dict):
        raise BadSchemaError(
            "File specification must be a mapping of file names to file specs, got {}".format(
                value
            )
        )

    if value.get("file_path"):
        # If the file spec was a list, this function will be called for each item. Just call this
        # function recursively to check each item.
        return validate_file_spec({"file": value}, rule_obj, path)

    for _, filespec in value.items():
        if isinstance(filespec, str):
            file_path = filespec
        elif isinstance(filespec, dict):
            valid = {"file_path", "content_type", "content_encoding", "form_field_name"}
            extra = set(filespec.keys()) - valid
            if extra:
                raise BadSchemaError(
                    f"Invalid extra keys passed to file upload block: {extra}"
                )

            try:
                file_path = filespec["file_path"]
            except KeyError as e:
                raise BadSchemaError(
                    "When using 'long form' file uplaod spec, the file_path must be present"
                ) from e
        else:
            raise BadSchemaError(
                "File specification must be a file path or a dictionary"
            )

        if not os.path.exists(file_path):
            if re.search(".*{.+}.*", file_path):
                logger.debug(
                    "Could not find file path, but it might be a format variable, so continuing"
                )
            else:
                raise BadSchemaError(
                    f"Path to file to upload '{file_path}' was not found"
                )

    return True


def raise_body_error(value, rule_obj, path):
    """Raise an error about the deprecated 'body' key"""

    msg = "The 'body' key has been replaced with 'json' in 1.0 to make it more in line with other blocks. see https://github.com/taverntesting/tavern/issues/495 for details."
    raise BadSchemaError(msg)


def retry_variable(value: int, rule_obj, path) -> bool:
    """Check retry variables"""

    int_variable(value, rule_obj, path)

    if isinstance(value, int):
        if value < 0:
            raise BadSchemaError("max_retries must be greater than 0")

    return True


def validate_http_method(value: str, rule_obj, path) -> bool:
    """Check http method"""

    if not isinstance(value, str):
        raise BadSchemaError("HTTP method should be a string")

    if value not in valid_http_methods:
        logger = get_pykwalify_logger("tavern.schemas.extensions")
        logger.debug(
            "Givern HTTP method '%s' was not one of %s - assuming it will be templated",
            value,
            valid_http_methods,
        )

    return True



import logging
import threading
from concurrent import futures
from typing import Any, Callable

import grpc
import helloworld_v1_precompiled_pb2 as helloworld_pb2_v1
import helloworld_v1_precompiled_pb2_grpc as helloworld_pb2_grpc_v1
import helloworld_v2_compiled_pb2 as helloworld_pb2_v2
import helloworld_v2_compiled_pb2_grpc as helloworld_pb2_grpc_v2
import helloworld_v3_reflected_pb2 as helloworld_pb2_v3
import helloworld_v3_reflected_pb2_grpc as helloworld_pb2_grpc_v3
from grpc_interceptor import ServerInterceptor
from grpc_interceptor.exceptions import GrpcException
from grpc_reflection.v1alpha import reflection


class GreeterV1(helloworld_pb2_grpc_v1.GreeterServicer):
    def SayHello(self, request, context):
        return helloworld_pb2_v1.HelloReply(message="Hello, %s!" % request.name)


class GreeterV2(helloworld_pb2_grpc_v2.GreeterServicer):
    def SayHello(self, request, context):
        return helloworld_pb2_v2.HelloReply(message="Hello, %s!" % request.name)


class GreeterV3(helloworld_pb2_grpc_v3.GreeterServicer):
    def SayHello(self, request, context):
        return helloworld_pb2_v3.HelloReply(message="Hello, %s!" % request.name)


class LoggingInterceptor(ServerInterceptor):
    def intercept(
        self,
        method: Callable,
        request_or_iterator: Any,
        context: grpc.ServicerContext,
        method_name: str,
    ) -> Any:
        logging.info(f"got request on {method_name}")

        try:
            return method(request_or_iterator, context)
        except GrpcException as e:
            logging.exception("error processing request")
            context.set_code(e.status_code)
            context.set_details(e.details)
            raise


def serve():
    interceptors = [LoggingInterceptor()]
    executor = futures.ThreadPoolExecutor(max_workers=10)

    # One server which exposes these two
    server = grpc.server(
        executor,
        interceptors=interceptors,
    )
    helloworld_pb2_grpc_v1.add_GreeterServicer_to_server(GreeterV1(), server)
    helloworld_pb2_grpc_v2.add_GreeterServicer_to_server(GreeterV2(), server)

    server.add_insecure_port("0.0.0.0:50051")
    server.start()

    # One server which exposes the V3 API and has reflection turned on
    reflecting_server = grpc.server(
        executor,
        interceptors=interceptors,
    )
    helloworld_pb2_grpc_v3.add_GreeterServicer_to_server(GreeterV3(), reflecting_server)
    service_names = (
        helloworld_pb2_v3.DESCRIPTOR.services_by_name["Greeter"].full_name,
        reflection.SERVICE_NAME,
    )
    reflection.enable_server_reflection(service_names, reflecting_server)
    reflecting_server.add_insecure_port("0.0.0.0:50052")
    reflecting_server.start()

    logging.info("Starting grpc server")
    event = threading.Event()
    event.wait()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    serve()



import dataclasses
import logging
import mimetypes
import os
from contextlib import ExitStack
from typing import Any, List, Optional, Tuple, Union

from tavern._core import exceptions
from tavern._core.dict_util import format_keys
from tavern._core.pytest.config import TestConfig

logger: logging.Logger = logging.getLogger(__name__)


@dataclasses.dataclass
class _Filespec:
    """A description of a file for a file upload, possibly as part of a multi part upload"""

    path: str
    content_type: Optional[str] = None
    content_encoding: Optional[str] = None
    form_field_name: Optional[str] = None


def _parse_filespec(filespec: Union[str, dict]) -> _Filespec:
    """
    Get configuration for uploading file

    Args:
        filespec: Can either be one of
            - A path to a file
            - A dict containing 'long' format, possibly including content type/encoding and the
                multipart 'name'

    Returns:
        The parsed file spec

    Raises:
        exceptions.BadSchemaError: If the file spec was invalid
    """
    if isinstance(filespec, str):
        return _Filespec(filespec)
    elif isinstance(filespec, dict):
        try:
            # The one required key
            path = filespec["file_path"]
        except KeyError as e:
            raise exceptions.BadSchemaError(
                "File spec dict did not contain the required 'file_path' key"
            ) from e

        return _Filespec(
            path,
            filespec.get("content_type"),
            filespec.get("content_encoding"),
            filespec.get("form_field_name"),
        )
    else:
        # Could remove, also done in schema check
        raise exceptions.BadSchemaError(
            "File specification must be a path or a dictionary"
        )


def guess_filespec(
    filespec: Union[str, dict], stack: ExitStack, test_block_config: TestConfig
) -> Tuple[List, Optional[str]]:
    """tries to guess the content type and encoding from a file.

    Args:
        test_block_config: config for test/stage
        stack: exit stack to add open files context to
        filespec: a string path to a file or a dictionary of the file path, content type, and encoding.

    Returns:
        A tuple of either length 2 (filename and file object), 3 (as before, with content type),
            or 4 (as before, with with content encoding). If a group name for the multipart upload
            was specified, this is also returned.

    Notes:
        If a 4-tuple is returned, the last element is a dictionary of headers to send to requests,
            _not_ the raw encoding value.
    """
    if not mimetypes.inited:
        mimetypes.init()

    parsed = _parse_filespec(filespec)

    filepath = format_keys(parsed.path, test_block_config.variables)
    filename = os.path.basename(filepath)

    # a 2-tuple ('filename', fileobj)
    file_spec = [
        filename,
        stack.enter_context(open(filepath, "rb")),
    ]

    # Try to guess as well, but don't override what the user specified
    guessed_content_type, guessed_encoding = mimetypes.guess_type(filepath)
    content_type = parsed.content_type or guessed_content_type
    encoding = parsed.content_encoding or guessed_encoding

    # If it doesn't have a mimetype, or can't guess it, don't
    # send the content type for the file
    if content_type:
        # a 3-tuple ('filename', fileobj, 'content_type')
        logger.debug("content_type for '%s' = '%s'", filename, content_type)
        file_spec.append(content_type)
        if encoding:
            # or a 4-tuple ('filename', fileobj, 'content_type', custom_headers)
            logger.debug("encoding for '%s' = '%s'", filename, encoding)
            # encoding is None for no encoding or the name of the
            # program used to encode (e.g. compress or gzip). The
            # encoding is suitable for use as a Content-Encoding header.
            file_spec.append({"Content-Encoding": encoding})

    return file_spec, parsed.form_field_name


def _parse_file_mapping(file_args, stack, test_block_config) -> dict:
    """Parses a simple mapping of uploads where each key is mapped to one form field name which has one file"""
    files_to_send = {}
    for key, filespec in file_args.items():
        file_spec, form_field_name = guess_filespec(filespec, stack, test_block_config)

        # If it's a dict then the key is used as the name, at least to maintain backwards compatability
        if form_field_name:
            logger.warning(
                f"Specified 'form_field_name' as '{form_field_name}' in file spec, but the file name was inferred to be '{key}' from the mapping - the form_field_name will be ignored"
            )

        files_to_send[key] = tuple(file_spec)
    return files_to_send


def _parse_file_list(file_args, stack, test_block_config) -> List:
    """Parses a case where there may be multiple files uploaded as part of one form field"""
    files_to_send: List[Any] = []
    for filespec in file_args:
        file_spec, form_field_name = guess_filespec(filespec, stack, test_block_config)

        if not form_field_name:
            raise exceptions.BadSchemaError(
                "If specifying a list of files to upload for a multi part upload, the 'form_field_name' key must also be specified for each file to upload"
            )

        files_to_send.append(
            (
                form_field_name,
                tuple(file_spec),
            )
        )

    return files_to_send


def get_file_arguments(
    request_args: dict, stack: ExitStack, test_block_config: TestConfig
) -> dict:
    """Get correct arguments for anything that should be passed as a file to
    requests

    Args:
        request_args: args passed to requests
        test_block_config: config for test
        stack: context stack to add file objects to so they're
            closed correctly after use

    Returns:
        mapping of 'files' block to pass directly to requests
    """

    files_to_send: Optional[Union[dict, List]] = None

    file_args = request_args.get("files")

    if isinstance(file_args, dict):
        files_to_send = _parse_file_mapping(file_args, stack, test_block_config)
    elif isinstance(file_args, list):
        files_to_send = _parse_file_list(file_args, stack, test_block_config)
    elif file_args is not None:
        raise exceptions.BadSchemaError(
            f"'files' key in a HTTP request can only be a dict or a list but was {type(file_args)}"
        )

    if files_to_send:
        return {"files": files_to_send}
    else:
        return {}



import logging
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, TypedDict, Union

import grpc
import proto.message
from google.protobuf import json_format
from grpc import StatusCode

from tavern._core import exceptions
from tavern._core.dict_util import check_expected_keys
from tavern._core.exceptions import TestFailError
from tavern._core.pytest.config import TestConfig
from tavern._core.schema.extensions import to_grpc_status
from tavern._plugins.grpc.client import GRPCClient
from tavern.response import BaseResponse

if TYPE_CHECKING:
    from tavern._plugins.grpc.request import WrappedFuture

logger: logging.Logger = logging.getLogger(__name__)


GRPCCode = Union[str, int, List[str], List[int]]


def _to_grpc_name(status: GRPCCode) -> Union[str, List[str]]:
    if isinstance(status, list):
        return [_to_grpc_name(s) for s in status]  # type:ignore

    if status_name := to_grpc_status(status):
        return status_name.upper()

    # This should have been verified before this
    raise exceptions.GRPCServiceException(f"unknown status code '{status}'")


class _GRPCExpected(TypedDict):
    """What the 'expected' block for a grpc response should contain"""

    status: GRPCCode
    details: Any
    body: Mapping


class GRPCResponse(BaseResponse):
    def __init__(
        self,
        client: GRPCClient,
        name: str,
        expected: Union[_GRPCExpected, Mapping],
        test_block_config: TestConfig,
    ) -> None:
        check_expected_keys({"body", "status", "details", "save"}, expected)
        super().__init__(name, expected, test_block_config)

        self._client = client

    def __str__(self):
        if self.response:
            return self.response.payload
        else:
            return "<Not run yet>"

    def _validate_block(self, blockname: str, block: Mapping) -> None:
        """Validate a block of the response

        Args:
            blockname: which part of the response is being checked
            block: The actual part being checked
        """
        try:
            expected_block = self.expected["body"] or {}
        except KeyError:
            expected_block = {}

        if isinstance(expected_block, dict):
            if expected_block.pop("$ext", None):
                logger.warning(
                    "$ext function found in block %s - this has been moved to verify_response_with block - see documentation",
                    blockname,
                )

        logger.debug("Validating response %s against %s", blockname, expected_block)

        test_strictness = self.test_block_config.strict
        block_strictness = test_strictness.option_for(blockname)
        self.recurse_check_key_match(expected_block, block, blockname, block_strictness)

    def verify(self, response: "WrappedFuture") -> Mapping:
        grpc_response = response.response

        logger.debug(f"grpc status code: {grpc_response.code()}")
        logger.debug(f"grpc details: {grpc_response.details()}")

        verify_status = [StatusCode.OK.name]
        if status := self.expected.get("status", None):
            verify_status = _to_grpc_name(status)  # type: ignore
            if not isinstance(verify_status, list):
                verify_status = [verify_status]

        if grpc_response.code().name not in verify_status:
            self._adderr(
                "expected status %s, but the actual response '%s'",
                verify_status,
                grpc_response.code().name,
            )

        if "details" in self.expected:
            verify_details = self.expected["details"]
            if verify_details != grpc_response.details():
                self._adderr(
                    "expected details '%s', but the actual response '%s'",
                    verify_details,
                    grpc_response.details(),
                )

        saved = self._handle_grpc_response(grpc_response, response, verify_status) or {}

        if self.errors:
            raise TestFailError(
                f"Test '{self.name:s}' failed:\n{self._str_errors():s}",
                failures=self.errors,
            )

        return saved

    def _handle_grpc_response(
        self,
        grpc_response: Union[grpc.Call, grpc.Future],
        response: "WrappedFuture",
        verify_status: List[str],
    ) -> Optional[Dict[str, Any]]:
        if grpc_response.code().name != "OK":
            # TODO: Should allow checking grpc RPC error details etc.
            logger.info(
                f"skipping body checking due to {grpc_response.code()} response"
            )
            return None

        if "body" in self.expected and verify_status != ["OK"]:
            self._adderr(
                "'body' was specified in response, but expected status code was not 'OK'"
            )
            return None

        _, output_type = self._client.get_method_types(response.service_name)
        result: proto.message.Message = grpc_response.result()

        if not isinstance(result, output_type):
            # Note: This is probably unexpected in some cases
            self._adderr(
                f"response from server ({type(response)}) was not the same type as expected from the registered definition ({output_type})"
            )
            return None

        json_result = json_format.MessageToDict(
            result,
            including_default_value_fields=True,
            preserving_proto_field_name=True,
        )

        if "body" in self.expected:
            expected_parsed = output_type()
            try:
                json_format.ParseDict(self.expected["body"], expected_parsed)
            except json_format.ParseError as e:
                self._adderr(f"response body was not in the right format: {e}", e=e)

            self._validate_block("json", json_result)
            self._maybe_run_validate_functions(json_result)

        saved: Dict[str, Any] = {}
        saved.update(self.maybe_get_save_values_from_save_block("body", json_result))
        saved.update(self.maybe_get_save_values_from_ext(json_result, self.expected))

        return saved



import warnings

# Shut up warnings caused by proto libraries
warnings.filterwarnings(
    "ignore", category=DeprecationWarning, module="pkg_resources", lineno=2804
)
warnings.filterwarnings(
    "ignore", category=DeprecationWarning, module="pkg_resources", lineno=2309
)
warnings.filterwarnings(
    "ignore", category=DeprecationWarning, module="pkg_resources", lineno=2870
)
warnings.filterwarnings(
    "ignore", category=DeprecationWarning, module="pkg_resources", lineno=2349
)
warnings.filterwarnings(
    "ignore", category=DeprecationWarning, module="pkg_resources", lineno=20
)
